#########################################################
# Description: Geometry tools relying on VTK classes
# Provide interfaces for decimation, smoothing and deformation
# vtkPolyData is required for processing. Thus converter is provided here.
# Usage: check out Test() method

# Author: Shaoting Zhang, shaoting@cs.rutgers.edu

#########################################################

import os
import re
import copy
import numpy
import Utility
import vtk
import ASMtools

def LoadVTKMesh(filename):
    reader = vtk.vtkPolyDataReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()
    
def LoadOffMesh(off_filename):
    vtk_filename = off_filename.replace('.off', '.vtk')
    Utility.off2vtk(off_filename, vtk_filename)
    return LoadVTKMesh(vtk_filename)
    
def DisplayVTKMesh(mesh):
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInput(mesh)
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetInterpolationToPhong()
    actor.GetProperty().SetAmbient(0.2)
    actor.GetProperty().SetDiffuse(0.7)
    actor.GetProperty().SetSpecular(0.6)
    actor.GetProperty().SetSpecularPower(100)
    actor.GetProperty().SetColor(0.8,0.8,1)
    ren1 = vtk.vtkRenderer()
    ren1.AddActor(actor)
    ren1.SetBackground(1, 1, 1)
    renWin = vtk.vtkRenderWindow()
    renWin.AddRenderer(ren1)
    renWin.SetSize(300, 300)
    iren = vtk.vtkRenderWindowInteractor()
    iren.SetRenderWindow(renWin)
    style = vtk.vtkInteractorStyleTrackballCamera()
    iren.SetInteractorStyle(style)
    iren.Initialize()
    iren.Start()

def Decimate(mesh,
TargetReduction = 0.5,
PreserveTopology = False,
Splitting = False,
BoundaryVertexDeletion = True):
    deci = vtk.vtkDecimatePro()
    deci.SetInput(mesh)
    deci.SetTargetReduction(TargetReduction)
    if PreserveTopology == True:
        deci.PreserveTopologyOn()
    else:
        deci.PreserveTopologyOff()
    if Splitting == True:
        deci.SplittingOn()
    else:
        deci.SplittingOff()
    if BoundaryVertexDeletion == True:
        deci.BoundaryVertexDeletionOn()
    else:
        deci.BoundaryVertexDeletionOff()
    deci.Update()
    return deci.GetOutput()
    
def SaveVTKMesh(mesh, filename):
    writer = vtk.vtkPolyDataWriter()
    writer.SetInput(mesh)
    writer.SetFileName(filename)
    writer.Write()
    
def Smooth(mesh,
NumberOfIterations = 200,
FeatureEdgeSmoothing = False,
BoundarySmoothing = True):
    smooth = vtk.vtkSmoothPolyDataFilter()
    smooth.SetInput(mesh)
    if FeatureEdgeSmoothing == True:
        smooth.FeatureEdgeSmoothingOn()
    else:
        smooth.FeatureEdgeSmoothingOff()
    if BoundarySmoothing == True:
        smooth.BoundarySmoothingOn()
    else:
        smooth.BoundarySmoothingOff()
    smooth.SetNumberOfIterations(NumberOfIterations)
    smooth.Update()
    return smooth.GetOutput()
    
# vtkThinPlateSplineTransform describes a nonlinear warp transform defined 
# by a set of source and target landmarks. Any point on the mesh close to a 
# source landmark will be moved to a place close to the corresponding target 
# landmark. The points in between are interpolated smoothly using Bookstein's 
# Thin Plate Spline algorithm.
def Deform(mesh, SourceLandmarks, TargetLandmarks):
    tps = vtk.vtkThinPlateSplineTransform()
    tps.SetSourceLandmarks(SourceLandmarks)
    tps.SetTargetLandmarks(TargetLandmarks)

    t1 = vtk.vtkGeneralTransform()
    t1.SetInput(tps)

    f11 = vtk.vtkTransformPolyDataFilter()
    f11.SetInput(mesh)
    f11.SetTransform(t1)
    f11. Update()
    
    return f11.GetOutput()
    
# convert list to vtkPoints
def List2vtkPoints(SL_list, TL_list):
    size = len(SL_list)
    SL = vtk.vtkPoints()
    SL.SetNumberOfPoints(size)
    TL = vtk.vtkPoints()
    TL.SetNumberOfPoints(size)
    for i in range(size):
        SL.SetPoint(i, SL_list[i])
        TL.SetPoint(i, TL_list[i])
        #SL.SetPoint(i, SL_list[i][0], SL_list[i][1], 0)
        #TL.SetPoint(i, TL_list[i][0], TL_list[i][1], 0)
    return SL, TL
    
# point set is used in ASM_3D. Thus a converter may be useful
def VTKMesh2PointSet(mesh):
    SaveVTKMesh(mesh, 'C:\\tmp.vtk')
    Utility.vtk2off('C:\\tmp.vtk', 'C:\\tmp.off')
    pointSet = ASMtools.LoadOneMesh('C:\\tmp.off')
    os.remove('C:\\tmp.vtk')
    os.remove('C:\\tmp.off')
    return pointSet

def Test():
    sphere = vtk.vtkSphereSource()
    sphere.SetRadius(1)
    sphere.SetCenter(0,0,0)
    sphere.Update()
    SaveVTKMesh(sphere.GetOutput(), 'C:\\temp\\sphere.vtk')
    
    mesh = LoadVTKMesh('C:\\temp\\sphere.vtk')
    #decimatedMesh = Decimate(mesh)
    smoothedMesh = Smooth(mesh)
    SL, TL = List2vtkPoints([[0,0,-1],[0,0,1]], [[0,0,-1],[0,0,1]])
    deformedMesh = Deform(mesh, SL, TL)
    #DisplayVTKMesh(mesh)
    #SaveVTKMesh(deformedMesh, 'C:\\temp\\sphere_deform.vtk')
    SaveVTKMesh(smoothedMesh, 'C:\\test\\smoothed_mesh.vtk')
    #mesh = LoadVTKMesh('C:\\test\\smoothed_mesh.vtk')
    #VTKMesh2PointSet(mesh)
    DisplayVTKMesh(smoothedMesh)
    